#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <netinet/in.h>

#include <liburing.h>
#include <liburing/io_uring.h>

#define CONNECTIONS_BACKLOG 10
#define BUFFER_COUNT 10
#define BUFFER_LEN 1024

enum {
    OP_ACCEPT = 1,
    OP_RECV = 2,
    OP_PROVIDE_BUFFERS = 3,
};

char bufs[BUFFER_COUNT][BUFFER_LEN] = {0};
int server_port = 5001;
int bgid = 1337;

int main(int argc, char *argv[]) {
    struct sockaddr_in serv_addr = {0}, client_addr = {0};
    struct io_uring_params params = {0};
    struct io_uring ring = {0};
    struct io_uring_sqe *sqe;
    struct io_uring_cqe *cqe;
    const int val = 1;
    socklen_t client_len = sizeof(client_addr);

    int sock_listen_fd = socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);
    setsockopt(sock_listen_fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));

    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(server_port);
    serv_addr.sin_addr.s_addr = INADDR_ANY;

    bind(sock_listen_fd, (struct sockaddr *) &serv_addr, sizeof(serv_addr));
    listen(sock_listen_fd, CONNECTIONS_BACKLOG);

    io_uring_queue_init_params(1024, &ring, &params);

    // Setup the buffers
    sqe = io_uring_get_sqe(&ring);
    io_uring_prep_provide_buffers(sqe, bufs, BUFFER_LEN, BUFFER_COUNT, bgid, 0);
    io_uring_submit(&ring);
    io_uring_wait_cqe(&ring, &cqe);
    io_uring_cqe_seen(&ring, cqe);

    // Submit an accept sqe
    sqe = io_uring_get_sqe(&ring);
    io_uring_prep_accept(sqe, sock_listen_fd, (struct sockaddr *) &client_addr, &client_len, 0);
    sqe->user_data = ((uint8_t) OP_ACCEPT & 0xFFu);

    fprintf(stdout, "Waiting for connections on port %d\n", server_port);
    fflush(stdout);

    // start event loop
    while (1) {
        io_uring_submit_and_wait(&ring, 1);
        unsigned head;
        unsigned count = 0;

        // go through all CQEs
        io_uring_for_each_cqe(&ring, head, cqe) {
            ++count;

            int type = cqe->user_data & 0xFF;

            if (type == OP_ACCEPT) {
                if (cqe->res >= 0) {
                    int fd = cqe->res;

                    sqe = io_uring_get_sqe(&ring);
                    io_uring_prep_recv(sqe, fd, NULL, BUFFER_LEN, 0);
                    io_uring_sqe_set_flags(sqe, IOSQE_BUFFER_SELECT);
                    sqe->buf_group = bgid;
                    sqe->user_data = ((int64_t) fd << 8) | (OP_RECV & 0xFF);

                    fprintf(stdout, "[CQE][ACCEPT] new connection, fd %d\n", fd);
                    fflush(stdout);
                }

                sqe = io_uring_get_sqe(&ring);
                io_uring_prep_accept(sqe, sock_listen_fd, (struct sockaddr *) &client_addr, &client_len, 0);
                sqe->user_data = (OP_ACCEPT & 0xFF);
            } else if (type == OP_RECV) {
                int fd = (cqe->user_data >> 8) & 0xFFFFFFFF;
                int bytes_read = cqe->res;

                fprintf(stdout, "[CQE][RECV] fd %d, cqe flags high: %d, cqe flags low: %d\n",
                        fd, cqe->flags >> 16, cqe->flags & 0xFFFF);
                fprintf(stdout, "[CQE][RECV] bytes: %d\n", bytes_read);
                fflush(stdout);

                if (bytes_read <= 0) {
                    shutdown(fd, SHUT_RDWR);
                } else {
                    int bid = cqe->flags >> IORING_CQE_BUFFER_SHIFT;

                    fprintf(stdout, "[CQE][RECV] msg: %s\n", bufs[bid]);
                    fflush(stdout);

                    for (uint16_t i = 0; i < BUFFER_COUNT; i++) {
                        fprintf(stdout, "[BUFFER][%d] %s\n", i, bufs[i]);
                    }
                    fflush(stdout);

                    sqe = io_uring_get_sqe(&ring);
                    io_uring_prep_provide_buffers(sqe, bufs, BUFFER_LEN, 1, bgid, bid);
                    sqe->user_data = (OP_PROVIDE_BUFFERS & 0xFF);

                    sqe = io_uring_get_sqe(&ring);
                    io_uring_prep_recv(sqe, fd, NULL, BUFFER_LEN, 0);
                    io_uring_sqe_set_flags(sqe, IOSQE_BUFFER_SELECT);
                    sqe->buf_group = bgid;
                    sqe->user_data = ((int64_t) fd << 8) | (OP_RECV & 0xFF);
                }
            } else {
                fprintf(stdout, "[CQE][OTHER] ignoring\n");
                fflush(stdout);
            }
        }

        io_uring_cq_advance(&ring, count);
    }
}
